package main

import (
	"context"
	"database/sql"
	"sync"
	"time"

	"git.joco.dk/sng/fermentord/internal/controllers"
	"git.joco.dk/sng/fermentord/internal/dal"
	"git.joco.dk/sng/fermentord/pkg/temperature"
	"github.com/getsentry/sentry-go"
	"github.com/rs/zerolog/log"
)

func migrateDB(config configuration) {
	migrator, err := dal.NewMigrator(config.DB.Local.DSN)
	if err != nil {
		sentry.CaptureException(err)
		log.Fatal().Err(err).Msg("Error initializing DB migrations")
	}
	migrator.Up()
	serr, dberr := migrator.Close()
	if serr != nil {
		sentry.CaptureException(serr)
		log.Fatal().Err(serr).Msg("DB migration source error")
	}
	if dberr != nil {
		sentry.CaptureException(dberr)
		log.Fatal().Err(dberr).Msg("DB migration database error")
	}
}

func persistData(ctx context.Context, wg *sync.WaitGroup, db *dal.DAL, chState <-chan controllers.ChamberState, chTemp <-chan temperature.TemperatureReading) {
	defer wg.Done()

	for {
		select {
		case state, ok := <-chState:
			if !ok {
				break
			}

			ctxTo, cancel := context.WithTimeout(ctx, 10*time.Second)
			err := db.InReadWriteTransaction(ctxTo, func(tx *sql.Tx) error {
				return dal.SaveChamberState(ctxTo, tx, convertChamberState(state))
			})
			if err != nil {
				sentry.CaptureException(err)
				log.Error().Err(err).Msg("Error persisting state change")
			}
			cancel()

		case t, ok := <-chTemp:
			if !ok {
				break
			}

			ctxTo, cancel := context.WithTimeout(ctx, 10*time.Second)
			err := db.InReadWriteTransaction(ctxTo, func(tx *sql.Tx) error {
				return dal.SaveTemperatureReading(ctx, tx, t.Time, t.Sensor, t.MilliDegrees)
			})
			if err != nil {
				sentry.CaptureException(err)
				log.Error().Err(err).Msg("Error persisting temperature reading")
			}
			cancel()

		case <-ctx.Done():
			return
		}
	}
}
