package main

import (
	"context"
	"fmt"
	"net/http"
	"sync"
	"time"

	"git.joco.dk/sng/fermentord/internal/api"
	"git.joco.dk/sng/fermentord/internal/controllers"
	"git.joco.dk/sng/fermentord/internal/dal"
	"git.joco.dk/sng/fermentord/internal/hw"
	"git.joco.dk/sng/fermentord/internal/metrics"
	"git.joco.dk/sng/fermentord/pkg/daemon"
	"git.joco.dk/sng/fermentord/pkg/temperature"
	"github.com/fsnotify/fsnotify"
	"github.com/getsentry/sentry-go"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

func mainLoop(ctx context.Context, wg *sync.WaitGroup, db *dal.DAL, config *configuration) {
	defer wg.Done()

	temperature.Initialize()

	// Controller
	chCtrlTemp := make(chan temperature.TemperatureReading, 10)
	defer close(chCtrlTemp)
	ctrl := controllers.NewChamberController("Chamber 1", config.ControllerConfig, chCtrlTemp)

	// Configuration reload
	viper.OnConfigChange(func(in fsnotify.Event) {
		reloadConfiguration(config, ctrl)
	})
	viper.WatchConfig()
	reloadConfiguration(config, ctrl)

	gpio, err := hw.NewGpio()
	if err != nil {
		sentry.CaptureException(err)
		log.Fatal().
			Err(err).
			Msg("Error initializing GPIO")
	}
	defer gpio.Close()

	chDbTemp := make(chan temperature.TemperatureReading, 10)
	chDbState := make(chan controllers.ChamberState, 10)
	defer close(chDbTemp)
	defer close(chDbState)

	wg.Add(3)
	go persistData(ctx, wg, db, chDbState, chDbTemp)
	go ctrl.Run(ctx, wg)
	go temperature.PollSensors(ctx, wg, 1*time.Second)

	for {
		select {
		case reading := <-temperature.C:
			chCtrlTemp <- reading
			chDbTemp <- reading

		case state := <-ctrl.C:
			switch state {
			case controllers.ChamberStateIdle:
				metrics.State.Set(metrics.MetricStateIdle)
				gpio.StopCooler()
				gpio.StopHeater()

			case controllers.ChamberStateCooling:
				metrics.State.Set(metrics.MetricStateCooling)
				gpio.StopHeater()
				gpio.StartCooler()

			case controllers.ChamberStateHeating:
				metrics.State.Set(metrics.MetricStateHeating)
				gpio.StopCooler()
				gpio.StartHeater()
			}

			chDbState <- state

		case <-ctx.Done():
			return
		}
	}
}

func main() {
	// Sentry
	err := sentry.Init(sentry.ClientOptions{
		Dsn:              "https://7278625538334140991ce433e0ad292f@sentry.joco.dk/24",
		TracesSampleRate: 0.01,
	})
	if err != nil {
		log.Fatal().
			Err(err).
			Msg("Failed to initialize Sentry.")
	}
	defer sentry.Flush(10 * time.Second)

	// Configuration
	config := loadConfiguration()

	// Database
	migrateDB(*config)

	var db *dal.DAL
	if db, err := dal.NewDAL(config.DB.Local.DSN); err != nil {
		sentry.CaptureException(err)
		log.Fatal().Err(err).Msg("Error connecting to DB")
	} else {
		defer db.Close()
	}

	// HTTP
	mux := http.NewServeMux()
	srv := &http.Server{
		Addr:    fmt.Sprintf(":%v", config.HTTP.Port),
		Handler: mux,
	}
	api := api.NewAPI(db)

	// Metrics
	mux.Handle("/metrics", promhttp.Handler())
	mux.HandleFunc("/health", api.HealthCheck)
	mux.HandleFunc("/temperatures", api.GetTemperatures)
	mux.HandleFunc("/state-changes", api.GetStateChanges)

	// Main
	ctxb := context.Background()
	ctx, shutdown := context.WithCancel(ctxb)
	wg := &sync.WaitGroup{}

	wg.Add(1)
	go mainLoop(ctx, wg, db, config)

	go srv.ListenAndServe()

	daemon.WaitForExitSignal()

	// Initiate graceful shutdown.
	wg.Add(1)
	go shutdownHTTP(ctxb, wg, srv)
	shutdown()
	wg.Wait()
}

func shutdownHTTP(ctx context.Context, wg *sync.WaitGroup, srv *http.Server) {
	defer wg.Done()
	ctx2, cancel := context.WithTimeout(ctx, 10*time.Second)
	defer cancel()
	if err := srv.Shutdown(ctx2); err != nil {
		sentry.CaptureException(err)
		log.Error().
			Err(err).
			Msg("Error shutting down HTTP server.")
	}
}
